from abc import ABC, abstractmethod
from random import choice
from typing import Optional

from colorama import Fore

from surv_ai.lib.knowledge_store.interfaces import Knowledge, KnowledgeStoreInterface
from surv_ai.lib.knowledge_store.local import LocalKnowledgeStore
from surv_ai.lib.llm.interfaces import LargeLanguageModelClientInterface


class BaseAgent(ABC):
    def __init__(
        self,
        client: LargeLanguageModelClientInterface,
        knowledge_store: Optional[KnowledgeStoreInterface] = None,
        n_knowledge_items_per_prompt: int = 5,
        name: Optional[str] = None,
        _hyperparameters: Optional[dict] = None,
    ):
        if knowledge_store is None:
            knowledge_store = LocalKnowledgeStore()

        if not name:
            name = self.__class__.__name__

        self.client = client
        self.knowledge_store = knowledge_store

        self.n_knowledge_items_per_prompt = n_knowledge_items_per_prompt

        self._hyperparameters = _hyperparameters or {}
        self.name = name

        self.color = choice(
            [
                Fore.BLUE,
                Fore.CYAN,
                Fore.GREEN,
                Fore.RED,
                Fore.MAGENTA,
                Fore.YELLOW,
                Fore.LIGHTGREEN_EX,
            ]
        )

    @abstractmethod
    async def prompt(self, statement: str, *args, **kwargs) -> str:
        ...

    def teach_text(self, input: str, source: Optional[str] = "User"):
        self.knowledge_store.add_text(f"{source}: {input}", source=source)

    def teach_knowledge(self, knowledge: Knowledge):
        self.knowledge_store.add_knowledge(knowledge)



import asyncio
from enum import Enum

from aiohttp import ClientSession

from surv_ai.lib.log import logger

from .interfaces import LargeLanguageModelClientInterface, Prompt


class AnthropicModel(str, Enum):
    CLAUDE_V1 = "claude-v1"


MODEL_TOKEN_LIMITS = {AnthropicModel.CLAUDE_V1: 100000}


class AnthropicClient(LargeLanguageModelClientInterface):
    def __init__(
        self,
        api_key: str,
    ):
        self.api_key = api_key

    async def _get_completion(
        self,
        session: ClientSession,
        prompt: Prompt,
        attempt=1,
        presence_penalty=0,
        frequency_penalty=0,
        temperature=1,
        top_p=1,
        max_tokens: int = 800,
        model=AnthropicModel.CLAUDE_V1,
        token_multiplier=1.6,
    ) -> str:
        for message in prompt.messages:
            if message.role in ["user", "system"]:
                message.role = "Human"
            elif message.role == "assistant":
                message.role = "Assistant"

        messages = (
            "\n\n".join(
                [f"{message.name if message.name else message.role}: {message.content}" for message in prompt.messages]
            )
            + "\n\nAssistant: "
        )

        response = None
        try:
            request = {
                "model": model,
                "prompt": messages,
                "temperature": temperature,
                "top_p": top_p,
                "max_tokens_to_sample": max_tokens,
                "stop_sequences": ["\n\nUser:"],
            }

            response = await session.post(
                "https://api.anthropic.com/v1/complete",
                json=request,
                headers={
                    "Content-Type": "application/json",
                    "x-api-key": f"{self.api_key}",
                },
            )

            try:
                response_body = await response.json()
            except Exception:
                response_body = await response.text()

            response.raise_for_status()
        except Exception as e:
            logger.log_exception(e)

            if not response or response.status == 429 or response.status == 502:
                await asyncio.sleep(0.5)

                if attempt < 5:
                    return await self._get_completion(session, prompt, attempt + 1)
            elif response.status == 400:
                if attempt < 5:
                    return await self._get_completion(
                        session,
                        prompt,
                        attempt + 1,
                        token_multiplier=token_multiplier - 0.2,
                    )
            else:
                raise Exception(
                    f"Call to GPT API failed with status {response.status}.",
                    response_body,
                )

        return response_body["completion"]

    async def get_completions(
        self,
        prompts: list[Prompt],
        presence_penalty=0,
        frequency_penalty=0,
        temperature=1,
        top_p=1,
        max_tokens: int = 800,
        model=AnthropicModel.CLAUDE_V1,
    ) -> list[str]:
        async with ClientSession() as session:
            return await asyncio.gather(
                *[
                    self._get_completion(
                        session,
                        prompt,
                        presence_penalty=presence_penalty,
                        frequency_penalty=frequency_penalty,
                        temperature=temperature,
                        top_p=top_p,
                        max_tokens=max_tokens,
                        model=model,
                    )
                    for prompt in prompts
                ],
            )



import asyncio
from enum import Enum

from aiohttp import ClientSession

from surv_ai.lib.log import logger

from .interfaces import LargeLanguageModelClientInterface, Prompt


class GPTModel(str, Enum):
    TURBO = "gpt-3.5-turbo"
    GPT4 = "gpt-4"


MODEL_TOKEN_LIMITS = {GPTModel.TURBO: 4096, GPTModel.GPT4: 32000}


class GPTClient(LargeLanguageModelClientInterface):
    def __init__(
        self,
        api_key: str,
    ):
        self.api_key = api_key

    async def _get_completion(
        self,
        session: ClientSession,
        prompt: Prompt,
        attempt=1,
        presence_penalty=0,
        frequency_penalty=0,
        temperature=1,
        top_p=1,
        max_tokens: int = 800,
        model=GPTModel.TURBO,
        token_multiplier=1.6,
    ) -> str:
        MAX_PROMPT_TOKENS = MODEL_TOKEN_LIMITS.get(model) - max_tokens
        messages = [
            {
                "role": message.role,
                "content": " ".join(message.content.split()),
            }
            for message in prompt.messages
        ]
        approximate_tokens = len(str(messages).split(" ")) * token_multiplier

        while approximate_tokens > MAX_PROMPT_TOKENS:
            if len(messages) == 1:
                raise Exception("Initial prompt is too long.")

            index_to_remove = None
            for index, _ in enumerate(messages):
                if index == 0:
                    continue

                index_to_remove = index
                break

            if index_to_remove:
                messages.pop(index_to_remove)

            approximate_tokens = len(str(messages).split(" ")) * 1.7

        response = None
        try:
            request = {
                "model": model,
                "messages": messages,
                "temperature": temperature,
                "top_p": top_p,
                "presence_penalty": presence_penalty,
                "frequency_penalty": frequency_penalty,
                "max_tokens": max_tokens,
            }

            response = await session.post(
                "https://api.openai.com/v1/chat/completions",
                json=request,
                headers={
                    "Content-Type": "application/json",
                    "Authorization": f"Bearer {self.api_key}",
                },
            )

            try:
                response_body = await response.json()
            except Exception:
                response_body = await response.text()

            response.raise_for_status()
        except Exception as e:
            logger.log_exception(e)
            if not response or response.status == 429 or response.status == 502:
                await asyncio.sleep(0.5)

                if attempt < 5:
                    return await self._get_completion(session, prompt, attempt + 1)
            elif response.status == 400:
                if attempt < 5:
                    return await self._get_completion(
                        session,
                        prompt,
                        attempt + 1,
                        token_multiplier=token_multiplier - 0.2,
                    )
            else:
                raise Exception(
                    f"Call to GPT API failed with status {response.status}.",
                    response_body,
                )

        return response_body["choices"][0]["message"]["content"]

    async def get_completions(
        self,
        prompts: list[Prompt],
        presence_penalty=0,
        frequency_penalty=0,
        temperature=1,
        top_p=1,
        max_tokens: int = 800,
        model=GPTModel.TURBO,
    ) -> list[str]:
        async with ClientSession() as session:
            return await asyncio.gather(
                *[
                    self._get_completion(
                        session,
                        prompt,
                        presence_penalty=presence_penalty,
                        frequency_penalty=frequency_penalty,
                        temperature=temperature,
                        top_p=top_p,
                        max_tokens=max_tokens,
                        model=model,
                    )
                    for prompt in prompts
                ],
            )



from re import Pattern
from typing import Protocol

from ..llm.interfaces import LargeLanguageModelClientInterface


class NoMemoriesFoundException(Exception):
    ...


class ToolInterface(Protocol):
    instruction: str
    command: Pattern

    def __init__(self, client: LargeLanguageModelClientInterface, *args, **kwargs):
        ...

    async def use(self, *args, **kwargs):
        ...


class ToolBeltInterface(Protocol):
    tools: list[ToolInterface]

    def __init__(
        self,
        client: LargeLanguageModelClientInterface,
        tools: list[ToolInterface],
    ):
        ...

    async def observe(self, query: str):
        ...



from surv_ai.lib.knowledge_store.interfaces import Knowledge
from surv_ai.lib.llm.interfaces import Prompt, PromptMessage
from surv_ai.lib.log import logger

from ..agent import BaseAgent
from ..interfaces import AgentInterface


class ReasoningAgent(BaseAgent, AgentInterface):
    def _get_completion_prompt_text(self, prompt: str):
        return f"""     
        You are a citizen who must vote on whether this statement is more likely to be true or false:

        {prompt}
           
        You must decide whether you think the statement is more likely to be true or false.

        You must cast your vote and participate in the civic process.

        The next two messages will be arguments in favor and against the above statement.

        You must decide which argument you are more persuaded by. You may not be undecided.
        """

    def _get_argument_prompt_text(self, prompt: str):
        return f"""     
        You are a citizen who must vote on whether this statement is more likely to be true or false:

        {prompt}
           
        You must decide whether you think the statement is more likely to be true or false.

        You must cast your vote and participate in the civic process.

        The next set of messages will be a series of articles that you can use to help you make your decision.

        In your response please include as many citations from your research as possible.

        Include both the publication title and the article title in your citations.

        You must make the best decision possible with the information you have. 
        """

    def _get_plan_prompt_text(self, prompt: str):
        return f"""     
        You are a decisive agent who is concerned with whether this statement is more likely to be true or false:

        {prompt}
           
        You must decide whether you think the statement is more likely to be true or false.

        The next set of messages will be a series of articles that you can use to help you make your decision.

        In your response please construct a plan for how you will approach the problem.

        Make sure your plan includes offering specific citations.

        You must make the best decision possible with the information you have. You cannot be undecided.
        """

    async def _get_plan(self, prompt: str, relevant_knowledge: list[Knowledge]):
        messages = [
            PromptMessage(
                role="system",
                content=self._get_plan_prompt_text(
                    prompt,
                ),
            ),
            *[
                PromptMessage(
                    content=knowledge.text,
                    role="user",
                    name=knowledge.source,
                )
                for knowledge in relevant_knowledge
            ],
            PromptMessage(
                role="assistant",
                content="""In order to determine whether the provided statement is more likely to be true or false,
                This is an outline of how I will think about the problem:
                """,
            ),
        ]

        return (await self.client.get_completions([Prompt(messages=messages)]))[0]

    async def _get_argument_in_favor(
        self,
        prompt: str,
        relevant_knowledge: list[Knowledge],
    ):
        messages = [
            PromptMessage(
                role="system",
                content=self._get_argument_prompt_text(
                    prompt,
                ),
            ),
            *[
                PromptMessage(
                    content=knowledge.text,
                    role="user",
                    name=knowledge.source,
                )
                for knowledge in relevant_knowledge
            ],
            PromptMessage(
                role="assistant",
                content=f"""
                Thus after looking at the relevant sources, I have come to the conclusion that the assertion
                "{prompt}" is more likely to be true. I will now explain my reasoning step by step while citing only the above sources:
                """,
            ),
        ]

        return (await self.client.get_completions([Prompt(messages=messages)]))[0]

    async def _get_argument_against(
        self,
        prompt: str,
        relevant_knowledge: list[Knowledge],
    ):
        messages = [
            PromptMessage(
                role="system",
                content=self._get_argument_prompt_text(
                    prompt,
                ),
            ),
            *[
                PromptMessage(
                    content=knowledge.text,
                    role="user",
                    name=knowledge.source,
                )
                for knowledge in relevant_knowledge
            ],
            PromptMessage(
                role="assistant",
                content=f"""
                Thus after looking at the relevant sources, I have come to the conclusion that the assertion
                "{prompt}" is more likely to be false. I will now explain my reasoning step by step while citing only the above sources:
                """,
            ),
        ]

        return (await self.client.get_completions([Prompt(messages=messages)]))[0]

    async def _build_completion_prompt(
        self,
        prompt: str,
    ) -> Prompt:
        relevant_knowledge = self.knowledge_store.recall_recent(
            n_knowledge_items=self.n_knowledge_items_per_prompt,
        )

        plan = await self._get_plan(prompt, relevant_knowledge)
        logger.log_internal(f"{self.name} plans: {plan}")

        argument_in_favor = await self._get_argument_in_favor(prompt, relevant_knowledge)
        argument_against = await self._get_argument_against(prompt, relevant_knowledge)

        messages = [
            PromptMessage(
                role="system",
                content=self._get_completion_prompt_text(
                    prompt,
                ),
            ),
            PromptMessage(
                role="assistant",
                content=argument_in_favor,
                name="Argument In Favor",
            ),
            PromptMessage(
                role="assistant",
                content=argument_against,
                name="Argument Against",
            ),
            PromptMessage(
                role="assistant",
                content=f"""My approach to answering the question will be:

                {plan}
                
                Now, after consider both perspectives I am more persuaded by the argument stating:
                """,
            ),
        ]

        return Prompt(messages=messages)

    async def prompt(self, statement: str, *args, **kwargs) -> str:
        prompt = await self._build_completion_prompt(statement, *args, **kwargs)

        response = (await self.client.get_completions([prompt], **self._hyperparameters))[0]

        return response



from surv_ai.lib.conversation.interfaces import ConversationInterface
from surv_ai.lib.llm.interfaces import Prompt, PromptMessage

from ..agent import BaseAgent


class BinaryAgent(BaseAgent):
    def _get_initial_prompt_text(self, assertion: str):
        return f"""        
        Your job is to determine if the user believes a statement is true or false.

        Here is the original assertion:
        
        "{assertion}"

        The next message will be the statement for the user.

        Your only options are "True," or "False".

        Please always respond with a single word.
        """

    async def _build_completion_prompt(self, conversation: ConversationInterface) -> str:
        assertion = self.knowledge_store.recall_recent(
            n_knowledge_items=1,
            include_sources=["Assertion"],
        )[0]

        self.messages = [
            PromptMessage(
                role="system",
                content=self._get_initial_prompt_text(assertion),
            ),
            PromptMessage(
                role="user",
                content=conversation.as_string(),
            ),
            PromptMessage(
                role="assistant",
                content="I believe the user thinks this statement is: ",
            ),
        ]

        return Prompt(messages=self.messages)

    async def prompt(self, statement: str, *args, **kwargs) -> str:
        prompt = await self._build_completion_prompt(statement, *args, **kwargs)

        response = (await self.client.get_completions([prompt], **self._hyperparameters))[0]

        return response



from typing import Optional

from .interfaces import Knowledge, KnowledgeStoreInterface


class LocalKnowledgeStore(KnowledgeStoreInterface):
    def __init__(self):
        self.knowledge: list[Knowledge] = []

    def add_text(self, input: str, source: Optional[str] = None):
        self.knowledge.append(Knowledge(text=input, source=source))

    def add_knowledge(self, knowledge: Knowledge):
        self.knowledge.append(knowledge)

    def filter_knowledge(
        self,
        include_sources: Optional[list[str]] = None,
        exclude_sources: Optional[list[str]] = None,
    ):
        knowledge = self.knowledge

        if include_sources:
            knowledge = [knowledge_item for knowledge_item in knowledge if knowledge_item.source in include_sources]

        if exclude_sources:
            knowledge = [knowledge_item for knowledge_item in knowledge if knowledge_item.source not in exclude_sources]

        return knowledge

    def recall_recent(
        self,
        n_knowledge_items=5,
        include_sources: Optional[list[str]] = None,
        exclude_sources: Optional[list[str]] = None,
    ) -> list[Knowledge]:
        return [knowledge_item for knowledge_item in self.filter_knowledge(include_sources, exclude_sources)][
            -n_knowledge_items:
        ]

    @staticmethod
    def knowledge_as_string(knowledge: list[Knowledge]) -> str:
        return "\n".join(f"{i + 1}. {m.text}" for i, m in enumerate(knowledge)) if knowledge else ""



from typing import Optional

from colorama import Fore

from surv_ai.lib.log import logger

from .interfaces import ChatMessage, ConversationInterface


class Conversation(ConversationInterface):
    def __init__(self):
        self.history: list[ChatMessage] = []

    def add(self, message: str, speaker: str, color: str = Fore.LIGHTBLACK_EX):
        logger.log_output(f"{speaker}: {message}", color)
        self.history.append(ChatMessage(text=message, speaker=speaker, color=color))

    def __iter__(self):
        return iter(self.history)

    def __len__(self):
        return len(self.history)

    def __getitem__(self, index):
        return self.history[index]

    def as_string(self, n_most_recent=5, exclude_speakers: Optional[list[str]] = None):
        if exclude_speakers is None:
            exclude_speakers = []

        return (
            "```\n"
            + "\n\n".join(
                [str(message) for message in self.history[-n_most_recent:] if message.speaker not in exclude_speakers]
            )
            + "\n```"
        )



import re

from surv_ai.lib.knowledge_store.interfaces import Knowledge
from surv_ai.lib.llm.interfaces import (
    LargeLanguageModelClientInterface,
    Prompt,
    PromptMessage,
)
from surv_ai.lib.log import logger

from .interfaces import NoMemoriesFoundException, ToolBeltInterface, ToolInterface


class ToolBelt(ToolBeltInterface):
    def __init__(
        self,
        client: LargeLanguageModelClientInterface,
        tools: list[ToolInterface],
    ):
        self.client = client
        self.tools = tools

    @staticmethod
    def tools_as_list(tools: list[ToolInterface]) -> str:
        return "\n".join(f"{i + 1}. {t.instruction}" for i, t in enumerate(tools)) if tools else ""

    def _get_tool_belt_prompt(self, original_prompt: str, base_knowledge: list[Knowledge]) -> str:
        return Prompt(
            messages=[
                PromptMessage(
                    role="system",
                    content=f"""A user will prompt you with a statement.

                    You should conduct research to assess this statement.
                    
                    These are the commands available to you:

                    {self.tools_as_list(self.tools)}

                    You MUST respond with a command.
                    """,
                ),
                *[
                    PromptMessage(
                        role="system",
                        content=f"This information may be relevant to your query: {knowledge.text}",
                    )
                    for knowledge in base_knowledge
                ],
                PromptMessage(
                    role="user",
                    content=f"""{original_prompt}""",
                ),
                PromptMessage(
                    role="assistant",
                    content="""
                    In order to research the user's prompt, I will execute the following command:
                    """,
                ),
            ]
        )

    async def inspect(
        self,
        original_prompt: str,
        base_knowledge: list[Knowledge],
        attempt=1,
    ):
        prompt = self._get_tool_belt_prompt(original_prompt, base_knowledge)

        response = (await self.client.get_completions([prompt], **{"temperature": 0.7}))[0].strip()

        for tool in self.tools:
            match = re.match(tool.command, response)

            if match:
                knowledge = []
                for args in match.groups():
                    logger.log_context(f"...Using tool: {response}...")

                    knowledge += await tool.use(original_prompt, args)

                    if not knowledge and attempt < 3:
                        knowledge = await self.inspect(original_prompt, attempt=attempt + 1)
                    elif not knowledge:
                        raise NoMemoriesFoundException()

                return knowledge

        return []



from typing import Protocol

from pydantic import BaseModel

from ..llm.interfaces import PromptMessage


class ChatMessage(BaseModel):
    text: str
    speaker: str
    color: str

    def __str__(self):
        return f'{self.speaker} said, "{self.text}"'


class ConversationInterface(Protocol):
    def __init__(self, *args, **kwargs):
        ...

    def add(self, message: PromptMessage):
        ...



import logging
from enum import IntEnum

from colorama import Fore, Style


class AgentLogLevel(IntEnum):
    EXCEPTION = 0
    ERROR = 1
    WARNING = 2

    CONTEXT = 3
    OUTPUT = 4
    INTERNAL = 5


class Logger:
    def __init__(self, name="surv_ai", log_level=AgentLogLevel.ERROR):
        self.name = name
        self.log_level = log_level

        self._logger = logging.getLogger(name)

    def set_log_level(self, log_level):
        self.log_level = log_level

        if log_level in [
            AgentLogLevel.CONTEXT,
            AgentLogLevel.OUTPUT,
            AgentLogLevel.INTERNAL,
        ]:
            self._logger.setLevel(logging.INFO)
        elif log_level in [
            AgentLogLevel.ERROR,
            AgentLogLevel.WARNING,
        ]:
            self._logger.setLevel(logging.WARNING)
        elif log_level == AgentLogLevel.EXCEPTION:
            self._logger.setLevel(logging.ERROR)

    def log_context(self, message: str, color: str = Fore.WHITE):
        if self.log_level >= AgentLogLevel.CONTEXT:
            self._logger.info(f"{Style.BRIGHT}{color}{message}{Style.RESET_ALL}")

    def log_output(self, message: str, color: str = Fore.WHITE):
        if self.log_level >= AgentLogLevel.OUTPUT:
            self._logger.info(f"{Style.BRIGHT}{color}{message}{Style.RESET_ALL}")

    def log_internal(self, message: str, color: str = Fore.LIGHTBLACK_EX):
        if self.log_level >= AgentLogLevel.INTERNAL:
            self._logger.info(f"{color}{Style.DIM}{message}{Style.RESET_ALL}")

    def log_exception(self, error):
        if self.log_level >= AgentLogLevel.EXCEPTION:
            self._logger.exception(error)

    def log_error(self, message: str):
        if self.log_level >= AgentLogLevel.ERROR:
            self._logger.error(f"{message}")

    def log_warning(self, message: str):
        if self.log_level >= AgentLogLevel.WARNING:
            self._logger.warning(f"{message}")


logger = Logger()



from .core.agent import BaseAgent  # noqa
from .core.agents.binary import BinaryAgent  # noqa
from .core.agents.reasoning import ReasoningAgent  # noqa
from .core.interfaces import DataPoint  # noqa
from .core.interfaces import (  # noqa
    AgentInterface,
    ModelInterface,
    SurveyInterface,
    SurveyParameter,
    SurveyResponse,
)
from .core.model import Model  # noqa
from .core.survey import Survey  # noqa
from .lib.conversation.conversation import Conversation  # noqa
from .lib.conversation.conversation import ConversationInterface  # noqa
from .lib.knowledge_store.interfaces import Knowledge  # noqa
from .lib.knowledge_store.interfaces import KnowledgeStoreInterface  # noqa
from .lib.knowledge_store.local import LocalKnowledgeStore  # noqa
from .lib.llm.anthropic import AnthropicClient  # noqa
from .lib.llm.gpt import GPTClient  # noqa
from .lib.llm.interfaces import LargeLanguageModelClientInterface  # noqa
from .lib.llm.interfaces import Prompt, PromptMessage  # noqa
from .lib.log import AgentLogLevel, logger  # noqa
from .lib.tools.query.google_custom_search import GoogleCustomSearchTool  # noqa
from .lib.tools.query.interfaces import ToolInterface  # noqa
from .lib.tools.query.wikipedia import WikipediaTool  # noqa
from .lib.tools.tool_belt import ToolBelt  # noqa



import asyncio
import re

from aiohttp import ClientSession
from bs4 import BeautifulSoup

from surv_ai.lib.knowledge_store.interfaces import Knowledge
from surv_ai.lib.llm.interfaces import (
    LargeLanguageModelClientInterface,
    Prompt,
    PromptMessage,
)
from surv_ai.lib.log import logger

from .interfaces import QueryToolInterface


class GoogleCustomSearchTool(QueryToolInterface):
    instruction = """
        `SEARCH(keywords)` - use keywords to search the web for additional information.
    """
    command = r"SEARCH\((.+)\)"

    _base_url = "https://www.googleapis.com/customsearch/v1/siterestrict"

    def __init__(
        self,
        llm_client: LargeLanguageModelClientInterface,
        google_api_key: str,
        google_search_engine_id: str,
        start_date=None,
        end_date=None,
        n_pages=1,
        max_percent_per_source=1.0,
        max_concurrency=10,
        only_include_sources=None,
    ):
        self.client = llm_client

        self.google_api_key = google_api_key
        self.google_search_engine_id = google_search_engine_id

        self.n_pages = n_pages

        self.start_date = start_date
        self.end_date = end_date

        self.max_percent_per_source = max_percent_per_source
        self.max_concurrency = max_concurrency

        self.only_include_sources = only_include_sources

    async def _search(self, session, query: str) -> list[str]:
        start = 1
        results = []

        seen_sources = {}

        while len(results) < self.n_pages:
            params = {
                "key": self.google_api_key,
                "cx": self.google_search_engine_id,
                "q": re.sub(r"[^A-Za-z0-9 ]+", "", query),
                "start": start,
                "num": 10,
            }

            if self.start_date:
                params["q"] += f" after:{self.start_date}"

            if self.end_date:
                params["q"] += f" before:{self.end_date}"

            async with session.get(self._base_url, params=params) as response:
                data = await response.json()

            try:
                new_records = data["items"]
            except Exception:
                logger.log_exception("Could not retrieve all articles.")

                return results

            if not new_records:
                break

            if self.max_percent_per_source:
                records_to_return = []

                for record in new_records:
                    if not seen_sources.get(record["displayLink"]):
                        seen_sources[record["displayLink"]] = 0

                    if self.only_include_sources:
                        compatible_source = False
                        for source in self.only_include_sources:
                            if source in record["displayLink"]:
                                compatible_source = True

                        if not compatible_source:
                            continue

                    if seen_sources[record["displayLink"]] >= (self.n_pages * self.max_percent_per_source):
                        continue

                    seen_sources[record["displayLink"]] += 1
                    records_to_return.append(record)

                new_records = records_to_return

            results += new_records
            start += 10

        return results

    async def _get_page_text(self, session: ClientSession, web_url: str) -> str:
        response = await session.get(
            web_url,
            headers={"User-Agent": "Mozilla/5.0"},
        )

        data = await response.text()

        soup = BeautifulSoup(data, "html.parser")

        results = []
        for paragraph in soup.find_all("p"):
            results.append(paragraph.text)

        return results

    async def _ingest_page_information(
        self,
        session,
        original_prompt: str,
        title: str,
        publication: str,
        web_url: str,
    ):
        paragraphs = await self._get_page_text(session, web_url)

        prompt = Prompt(
            messages=[
                PromptMessage(
                    content=f"""A user has asked you some questions:
                    
                    {original_prompt}

                    All subsequent messages will be paragraphs from a {publication} article titled "{title}."

                    Your job is to extract any useful information that might help someone answer these questions.

                    Remember to include as much data and concrete examples as possible from the article.

                    For each useful piece of information you extract, please state why it relates to the original questions.
                    """,
                    role="system",
                ),
                *[PromptMessage(content=paragraph, role="user", name="Article") for paragraph in paragraphs],
                PromptMessage(
                    role="assistant",
                    content=f"I have read the Wikpedia article entitled {title} and some useful information is:",
                ),
            ],
        )
        response = await self.client.get_completions([prompt])

        return response[0]

    async def _ingest_pages(
        self,
        session: ClientSession,
        original_prompt: str,
        result: dict,
    ):
        try:
            metatags = result["pagemap"]["metatags"][0]

            publication = metatags.get("og:site_name", result["displayLink"])
            title = metatags.get("og:title", result["title"])

            logger.log_context(f"......Retrieving {publication} article with title {title}......")
            page_summary = await self._ingest_page_information(
                session,
                original_prompt,
                title,
                publication,
                result["link"],
            )

            return Knowledge(
                text=f'{publication} article entitled "{title}": {page_summary}',
                source=result["link"],
            )
        except Exception as e:
            logger.log_exception(e)
            return None

    async def use(
        self,
        original_prompt: str,
        search_query: str,
    ) -> list[Knowledge]:
        async with ClientSession() as session:
            search_results = await self._search(session, search_query)

            if len(search_results) == 0:
                return []

            response = []

            while len(response) < len(search_results[: self.n_pages]):
                results_to_fetch = min(
                    self.max_concurrency,
                    len(search_results[: self.n_pages]) - len(response),
                )

                response += await asyncio.gather(
                    *[
                        self._ingest_pages(session, original_prompt, result)
                        for result in search_results[len(response) : len(response) + results_to_fetch]
                    ]
                )

            return [r for r in response if r]



from typing import Any, Optional, Protocol

from pydantic import BaseModel
from typing_extensions import TypedDict, Unpack

from surv_ai.lib.knowledge_store.interfaces import Knowledge, KnowledgeStoreInterface
from surv_ai.lib.llm.interfaces import LargeLanguageModelClientInterface
from surv_ai.lib.tools.interfaces import ToolBeltInterface


class AgentInterface(Protocol):
    def __init__(
        self,
        client: LargeLanguageModelClientInterface,
        knowledge_store: Optional[KnowledgeStoreInterface],
        tool_belt: Optional[list[ToolBeltInterface]],
    ):
        ...

    async def prompt(self, input: str) -> str:
        ...

    def teach_text(self, input: str, source: Optional[str] = "User"):
        ...

    def teach_knowledge(self, knowledge: Knowledge):
        ...


class SurveyResponse(BaseModel):
    in_favor: int
    against: int
    undecided: int
    error: int

    percent_in_favor: float
    uncertainty: float


class SurveyKwargs(TypedDict):
    client: LargeLanguageModelClientInterface
    tool_belt: ToolBeltInterface
    n_agents: int
    max_concurrency: int
    max_knowledge_per_agent: int
    base_knowledge: Optional[list[Knowledge]]


class SurveyInterface(Protocol):
    def __init__(self, **kwargs: Unpack[SurveyKwargs]):
        ...

    async def conduct(self, hypothesis: str) -> SurveyResponse:
        ...


class SurveyParameter(BaseModel):
    kwargs: dict
    independent_variable: Any


class DataPoint(BaseModel):
    response: SurveyResponse
    parameter: SurveyParameter


class ModelInterface(Protocol):
    def __init__(
        self,
        survey_class: type[SurveyInterface],
        parameters: list[SurveyParameter],
        max_concurrency: int = 1,
    ):
        ...

    async def build(self, hypothesis: str, *parameter_set: Unpack[dict]) -> list[DataPoint]:
        ...
